/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung - Bibliothek
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.lib;

import java.io.Serializable;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;

/**
 * Fasst alle Informationen zusammen, die notwendig sind, um einen Parameter zu
 * beschreiben.
 *
 * @author BitCtrl Systems GmbH, Schumann, Peuker
 */
public class ParameterInfo {

	/**
	 * Serialisierbare Ausprägung der Klasse zur Übertragung per
	 * Datenverteiler-Attributgruppe.
	 *
	 * @author peuker
	 */
	public static final class SerializableParameterInfo implements Serializable {

		/** die Versions-ID für die Serialisierung. */
		private static final long serialVersionUID = 1L;

		/** die ID des Systemobjekts. */
		private final long objId;

		/** die ID des Typobjekts. */
		private final long typId;

		/** die ID der Attributgruppe. */
		private final long atgId;

		/** die Simulationsvariante. */
		private short sim;

		/**
		 * Konstruktor.
		 *
		 * @param info
		 *            das Objekt, das in eine serialisierbare Form gewandelt
		 *            werden soll.
		 * @param verbindung
		 *            die Verbindung zum ermitteln datenverteilerspezifischer
		 *            Informationen
		 */
		public SerializableParameterInfo(final ParameterInfo info, final ClientDavInterface verbindung) {
			if (null != info.getObjekt()) {
				objId = info.getObjekt().getId();
			} else {
				objId = 0;
			}
			if (null != info.getTyp()) {
				typId = info.getTyp().getId();
			} else {
				typId = 0;
			}
			if (null != info.getAtg()) {
				atgId = info.getAtg().getId();
			} else {
				atgId = 0;
			}
			sim = info.getSim();
			if (sim == DataDescription.NO_SIMULATION_VARIANT_SET) {
				sim = verbindung.getClientDavParameters().getSimulationVariant();
			}
		}

		/**
		 * erzeugt aus dem Objekt, die "Normalform", die Referenzen auf die
		 * Elemente aus dem Datenmodell der Konfiguration des Datenverteilers
		 * enthält.
		 *
		 * @param verbindung
		 *            die verwendete Datenverteilerverbindung
		 * @return das Objekt
		 */
		public ParameterInfo erzeugeParameterInfo(final ClientDavInterface verbindung) {
			final DataModel model = verbindung.getDataModel();
			return new ParameterInfo(model.getObject(objId), (SystemObjectType) model.getObject(typId),
					(AttributeGroup) model.getObject(atgId), sim);
		}
	}

	/** das Systemobjekt, an dem die Parameter enthalten sein sollen. */
	private final SystemObject objekt;

	/** der Typ des Objekts, für den ein Parameter definiert werden soll. */
	private final SystemObjectType typ;

	/** die Attributgruppe, die den Parameterdatensatz beschreibt. */
	private final AttributeGroup atg;

	/** die Simulationsvariante, für die ein Parameter ermittelt werden soll. */
	private final short sim;

	/**
	 * Konstruiert eine Parameterbeschreibung für einen (lokalen) Parameter im
	 * Normalbetrieb (Simulationsvariante = 0).
	 *
	 * @param obj
	 *            Zu parametrierendes Systemobjekt
	 * @param atg
	 *            Die parametrierende Attributgruppe
	 */
	public ParameterInfo(final SystemObject obj, final AttributeGroup atg) {
		this(obj, obj.getType(), atg, DataDescription.NO_SIMULATION_VARIANT_SET);
	}

	/**
	 * Konstruiert eine Parameterbeschreibung für einen (lokalen) Parameter.
	 *
	 * @param obj
	 *            Zu parametrierendes Systemobjekt
	 * @param atg
	 *            Die parametrierende Attributgruppe
	 * @param sim
	 *            Simulationsvariante des Parameters
	 */
	public ParameterInfo(final SystemObject obj, final AttributeGroup atg, final short sim) {
		this(obj, obj.getType(), atg, sim);
	}

	/**
	 * Konstruiert eine Parameterbeschreibung für einen globalen Parameter im
	 * Normalbetrieb (Simulationsvariante = 0).
	 *
	 * @param obj
	 *            Ein Systemobjekt in der definierten Hierarchie, für dessen
	 *            Nachfolger der Parameter gelten soll
	 * @param typ
	 *            Objekttyp für den der Parameterwert gelten soll
	 * @param atg
	 *            Die parametrierende Attributgruppe
	 */
	public ParameterInfo(final SystemObject obj, final SystemObjectType typ, final AttributeGroup atg) {
		this(obj, typ, atg, DataDescription.NO_SIMULATION_VARIANT_SET);
	}

	/**
	 * Konstruiert eine Parameterbeschreibung für einen globalen Parameter im
	 * Normalbetrieb (Simulationsvariante = 0).
	 *
	 * @param obj
	 *            Ein Systemobjekt in der definierten Hierarchie, für dessen
	 *            Nachfolger der Parameter gelten soll
	 * @param typ
	 *            Objekttyp für den der Parameterwert gelten soll
	 * @param atg
	 *            Die parametrierende Attributgruppe
	 * @param sim
	 *            Simulationsvariante des Parameters
	 */
	public ParameterInfo(final SystemObject obj, final SystemObjectType typ, final AttributeGroup atg,
			final short sim) {
		if (obj == null) {
			throw new IllegalArgumentException(
					"Es wurde kein Objekt zum Bestimmen von Parameterinformationen angegeben");
		}
		objekt = obj;
		if (typ == null) {
			this.typ = objekt.getType();
		} else {
			this.typ = typ;
		}
		this.atg = atg;
		this.sim = sim;
	}

	/**
	 * {@inheritDoc}.
	 *
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(final Object vergleichsObjekt) {
		if (this == vergleichsObjekt) {
			return true;
		}
		if (null == vergleichsObjekt) {
			return false;
		}
		if (getClass() != vergleichsObjekt.getClass()) {
			return false;
		}
		final ParameterInfo other = (ParameterInfo) vergleichsObjekt;
		if (null == atg) {
			if (null != other.atg) {
				return false;
			}
		} else if (!atg.equals(other.atg)) {
			return false;
		}
		if (null == objekt) {
			if (null != other.objekt) {
				return false;
			}
		} else if (!objekt.equals(other.objekt)) {
			return false;
		}
		if (sim != other.sim) {
			return false;
		}
		if (null == typ) {
			if (null != other.typ) {
				return false;
			}
		} else if (!typ.equals(other.typ)) {
			return false;
		}
		return true;
	}

	/**
	 * gibt die parametrierende Attributgruppe zurück.
	 *
	 * @return die Attributgruppe
	 */
	public AttributeGroup getAtg() {
		return atg;
	}

	/**
	 * gibt das parametrierte Systemobjekt zurück.
	 *
	 * @return das Systemobjekt
	 */
	public SystemObject getObjekt() {
		return objekt;
	}

	/**
	 * liefert die Serialisierte Form des Objekts, zur Übertragung über eine
	 * Datenverteiler-Attributgruppe.
	 *
	 * @param verbindung
	 *            die Verbindung zum ermitteln datenverteilerspezifischer
	 *            Informationen
	 *
	 * @return das serialisierbare Objekt
	 */
	SerializableParameterInfo getSerializableParameterInfo(final ClientDavInterface verbindung) {
		return new SerializableParameterInfo(this, verbindung);
	}

	/**
	 * gibt die Simulationsvariante des Parameters zurück.
	 *
	 * @return die Simulationsvariante
	 */
	public short getSim() {
		return sim;
	}

	/**
	 * gibt den Typ des parametrierten Systemobjekts zurück.
	 *
	 * @return ein Objekttyp
	 */
	public SystemObjectType getTyp() {
		return typ;
	}

	/** {@inheritDoc} */
	@Override
	public int hashCode() {
		return objekt.hashCode();
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		final StringBuffer buffer = new StringBuffer(100);
		buffer.append('[');
		buffer.append(objekt);
		buffer.append(',');
		buffer.append(typ);
		buffer.append(',');
		buffer.append(atg);
		buffer.append(',');
		buffer.append(sim);
		buffer.append(']');

		return buffer.toString();
	}
}
