/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung - Bibliothek
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.lib;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die Repräsentation einer Anfrage an die Parametrierung.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class ParametrierungsAnfrage {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** Maximale Wartezeit auf eine Antwort der Parametrierung (in ms). */
	private static final int MAXIMALE_WARTEZEIT = 600000;

	/** der globaler Zähler für die ID der Anfrage. */
	private static int globaleId;

	/** die Parameter der Anfrage. */
	private final Serializable parameter;

	/** der Type der Anfrage. */
	private final ParametrierungsAnfrageTyp typ;

	/** das ermittelte Ergebnis. */
	private Serializable ergebnis;

	/** die ID der Anfrage. */
	private final int id;

	/** der Zeitpunkt der Anfrage. */
	private long anfrageZeit;

	/** Lock zur Synchronisation der Abfrage. */
	private final Object lock = new Object();

	/**
	 * Konstruktor, erzeugt eine {@link ParametrierungsAnfrage} zur Steuerung
	 * der Parametrierung.
	 *
	 * @param id
	 *            die ID der Anfrage, unterstützt wird nur -1 für Backup
	 */
	ParametrierungsAnfrage(final int id) {

		this.typ = ParametrierungsAnfrageTyp.PARAMETER;
		this.parameter = null;
		this.id = id;
		ParametrierungsAnfrage.LOGGER.finer("Erstelle Anfrage: ", new Object[] { typ, parameter, id });
	}

	/**
	 * Konstruktor, erzeugt eine {@link ParametrierungsAnfrage}.
	 *
	 * @param typ
	 *            der Typ der Anfrage
	 * @param parameter
	 *            die Parameter der Anfrage
	 */
	ParametrierungsAnfrage(final ParametrierungsAnfrageTyp typ, final Serializable parameter) {

		this.typ = typ;
		this.parameter = parameter;
		id = erzeugeId();

		ParametrierungsAnfrage.LOGGER.finer("Erstelle Anfrage: ", new Object[] { typ, parameter, id });
	}

	/**
	 * ermittelt die nächste ID für eine Anfrage.
	 *
	 * @return die ID
	 */
	private synchronized int erzeugeId() {
		ParametrierungsAnfrage.globaleId++;
		return ParametrierungsAnfrage.globaleId;
	}

	/**
	 * liefert die ID der Anfrage.
	 *
	 * @return die ID
	 */
	long getId() {
		return id;
	}

	/**
	 * setzt das Ergebnis der Anfrage.
	 *
	 * @param data
	 *            das Ergebnis
	 */
	void setErgebnis(final Data data) {

		ParametrierungsAnfrage.LOGGER.fine("Setze Ergebnis für Anfrage: " + id);

		try {
			if (null != data) {

				final byte[] bytes = data.getArray(ParametrierungsAnfrageDefs.ATT_NAME_DATEN).asUnscaledArray()
						.getByteArray();
				final ByteArrayInputStream input = new ByteArrayInputStream(bytes);
				final ObjectInputStream oi = new ObjectInputStream(input);
				ergebnis = (Serializable) oi.readObject();
			}
		} catch (final IOException e) {
			throw new IllegalStateException("Fehler beim einlesen der empfangenen Daten", e);
		} catch (final ClassNotFoundException e) {
			throw new IllegalStateException("Fehler beim einlesen der empfangenen Daten", e);
		} finally {
			synchronized (lock) {
				lock.notifyAll();
			}
		}
	}

	/**
	 * versendet die Anfrage und wartet, bis ein Ergebnis ermittelt wurde.<br>
	 * Wird innerhalb der erwarteten Zeit (20 Sekunden) kein Ergebnis empfangen,
	 * wird eine entsprechende Exception geworfen.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 * @param anfrageDesc
	 *            die Datenbeschreibung für eine Parameteranfrage
	 * @return das Ergebnis der Anfrage
	 * @throws ParameterClientException
	 *             die Anfrage konnte nicht erfolgreich ausgeführt werden
	 */
	Serializable warteAufErgebnis(final ClientDavInterface verbindung, final DataDescription anfrageDesc)
			throws ParameterClientException {

		ParametrierungsAnfrage.LOGGER.fine("Warte auf Ergebnis für Anfrage: " + id);

		final Data data = verbindung.createData(anfrageDesc.getAttributeGroup());
		data.getReferenceValue(ParametrierungsAnfrageDefs.ATT_NAME_ABSENDER)
				.setSystemObject(verbindung.getLocalApplicationObject());
		data.getUnscaledValue(ParametrierungsAnfrageDefs.ATT_NAME_ANFRAGE_INDEX).set(id);
		data.getUnscaledValue(ParametrierungsAnfrageDefs.ATT_NAME_ANFRAGE_TYP).set(typ.getCode());
		try {
			final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
			final ObjectOutputStream os = new ObjectOutputStream(byteStream);
			os.writeObject(parameter);
			final byte[] bytes = byteStream.toByteArray();
			final Data.Array array = data.getArray(ParametrierungsAnfrageDefs.ATT_NAME_DATEN);
			array.setLength(bytes.length);
			for (int idx = 0; idx < bytes.length; idx++) {
				array.getUnscaledValue(idx).set(bytes[idx]);
			}
		} catch (final IOException e) {
			throw new RuntimeException("Fehler bei der Ausgabe des Ergebnisses", e);
		}

		try {
			anfrageZeit = System.currentTimeMillis();
			ParametrierungsAnfrage.LOGGER.fine("Sende Anfrage: " + id);
			verbindung.sendData(
					new ResultData(verbindung.getLocalConfigurationAuthority(), anfrageDesc, anfrageZeit, data));
		} catch (final DataNotSubscribedException e) {
			throw new ParameterClientException("Die Anfrage konnte nicht versendet werden", e);
		} catch (final SendSubscriptionNotConfirmed e) {
			throw new ParameterClientException("Die Parametrierung empfängt keine Anfragen für die Simulationsvariante "
					+ verbindung.getClientDavParameters().getSimulationVariant(), e);
		}

		synchronized (lock) {
			try {
				lock.wait(ParametrierungsAnfrage.MAXIMALE_WARTEZEIT);
			} catch (final InterruptedException e) {
				Debug.getLogger().error("Anfrageschleife unterbrochen", e);
			}
		}

		if (null == ergebnis) {
			throw new ParameterClientException("Keine Antwort auf eine Parameteranfrage innerhalb von "
					+ ParametrierungsAnfrage.MAXIMALE_WARTEZEIT / 1000 + " Sekunden");
		}
		return ergebnis;
	}
}
